/* 
 * This is a subpixel rendering filter. It accepts a stream of raw PPM files
 * on stdin and for each such image, it first RGB decimates the image
 * and then invokes a user specified command line on the result, typically
 * to display it. The specified command line should accept a raw PPM file
 * on its stdin.
 * 
 *
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */
#include <X11/Xlib.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>

// X11 variables - these are global mainly for the alarm handler
Display *dpy;
Window win;
Window mwin;
Screen *scr;
Pixmap dest;
char *window_id;
Atom NEXT, PAGE, DONE;
XSetWindowAttributes attributes;
XWindowAttributes attributes_return;
XEvent event;

// alarm timeout
#define TIMEOUT 5

// other global variables
// used by the matrix renderer
double asum[3] = {9, 9, 9};
double a[3][9] = { {1, 1, 1, 1, 1, 1, 1, 1, 1},
		   {1, 1, 1, 1, 1, 1, 1, 1, 1},
		   {1, 1, 1, 1, 1, 1, 1, 1, 1} };



void usage(char *name) {
  fprintf(stdout, "%s accepts a stream of ppm files on STDIN which are subsequently\n", name);
  fprintf(stdout, "        rgb decimated and piped into a shell command as a reduced ppm.\n\n");
  fprintf(stdout, "usage: %s [spr|spr2|aa|mat xxx] command\n", name);
  fprintf(stdout, "       where command is the command line to execute for each decimated ppm.\n");
  fprintf(stdout, "       if the option mat is selected, then xxx must be a sequence of 27 numbers.\n");
  exit(0);
}

// pi: RGB RGB RGB
// po: rgb rgb rgb
// where r = (R1 + R2 + R3)/3
//       g = (G1 + G2 + G3)/3
//       b = (B1 + B2 + B3)/3
// this is the basic full pixel antialiasing, where
// each pixel is the average of itself and its two neighbours
void render_subpixels_aa(char *i, char *o, int w) {
  int k;
  unsigned char *pi, *po;
  unsigned char g,m,d;

  pi = i;
  po = o;
  for(k = 0; k < w; k++) {
    po[0] = (pi[0] / 3) + (pi[3] / 3) + (pi[6] / 3);
    po[1] = (pi[1] / 3) + (pi[4] / 3) + (pi[7] / 3);
    po[2] = (pi[2] / 3) + (pi[5] / 3) + (pi[8] / 3);
    pi += 9;
    po += 3;
  }
}

// pi: RGB RGB RGB
// po: rgb 
// where r = (R1 + G1 + B1)/3
//       g = (R2 + G2 + B2)/3
//       b = (R3 + G3 + B3)/3
// this is the basic subpixel rendering, where each full pixel
// is represented by a single color subpixel
void render_subpixels_spr(char *i, char *o, int w) {
  int k;
  unsigned char *pi, *po;
  unsigned char g,m,d;

  pi = i;
  po = o;
  for(k = 0; k < w; k++) {
    po[0] = (pi[0] / 3) + (pi[1] / 3) + (pi[2] / 3);
    po[1] = (pi[3] / 3) + (pi[4] / 3) + (pi[5] / 3);
    po[2] = (pi[6] / 3) + (pi[7] / 3) + (pi[8] / 3);
    pi += 9;
    po += 3;
  }
}

// pi: RGB RGB RGB
// po: rgb 
// where r = (B0 + 2R1 + 3G1 + 2B1 + R2)/9
//       g = (B1 + 2R2 + 3G2 + 2B2 + R3)/9
//       b = (B2 + 2R3 + 3G3 + 2B3 + R4)/9
void render_subpixels_spr2(char *i, char *o, int w) {
  int k;
  unsigned char *pi, *po;
  unsigned char g,m,d;

  pi = i;
  po = o;
  for(k = 0; k < w; k++) {
    po[0] = (pi[-1] / 9) + 2*(pi[0] / 9) + 3*(pi[1] / 9) + 2*(pi[2] / 9) + (pi[3] / 9);
    po[1] =  (pi[2] / 9) + 2*(pi[3] / 9) + 3*(pi[4] / 9) + 2*(pi[5] / 9) + (pi[6] / 9);
    po[2] =  (pi[5] / 9) + 2*(pi[6] / 9) + 3*(pi[7] / 9) + 2*(pi[8] / 9) + (pi[9] / 9);
    pi += 9;
    po += 3;
  }
}

// pi: RGB RGB RGB
// po: rgb 
// where r = (a11R1 + a12G1 + a13B1 + 
//            a14R2 + a15G2 + a16B2 + 
//            a17R3 + a18G3 + a19B3)/\sum_j a1j
//       g = (a21R1 + a22G1 + a23B1 + 
//            a24R2 + a25G2 + a26B2 + 
//            a27R3 + a28G3 + a29B3)/\sum_j a2j
//       b = (a31R1 + a32G1 + a33B1 + 
//            a34R2 + a35G2 + a36B2 + 
//            a37R3 + a38G3 + a39B3)/\sum_j a3j
// this is the fully general matrix filter. It requires a 3x9 matrix a whose
// entries can be found empirically. The matrix entries are global variables
void render_subpixels_matrix(char *i, char *o, int w) {
  int k, l, n;
  unsigned char *pi, *po;
  unsigned char g,m,d;

  pi = i;
  po = o;
  for(k = 0; k < w; k++) {
    for(n = 0; n < 3; n++) {
      po[n] = 0;
      for(l = 0; l < 9; l++) {
	if( a[n][l] > 0.01 ) { // tolerance 0.01
	  // slow yes, but otherwise we'll overflow the variable
	  po[n] += a[n][l] * (pi[l] / asum[n]); 
	}
      }
    }
    pi += 9;
    po += 3;
  }
}

// this will bomb if first chars are not P6\n
int pipe_ppm(FILE *COMMAND, void (*render_subpixels)()) {
  char b[1024];
  char *irasterbuf, *orasterbuf;
  int n, c, i;
  int width, w3, sw, sw3, height, colors;

  // check for magic number P6\n
  if( fgets(b, 3, stdin) == NULL ) {
    return 0; // failure, probably eof
  }
  if( strncmp(b, "P6", 2) != 0 ) {
    fprintf(stderr,"subpixelize: P6 error [%s]\n", b);
    return 0; // failure
  }

  // skip white space and comments
  // if it starts with a digit, its the
  // line we're looking for
  if( fgets(b, 1023, stdin) == NULL ) {
    return 0; // failure, probably eof
  }
  while( !isdigit(b[0]) ) {
    if( fgets(b, 1023, stdin) == NULL ) {
      return 0; // failure
    }
  }

  // get height and width of image
  sscanf(b, "%u%*[ \t]%u", &width, &height);
  if( !(width * height > 0) ) {
    fprintf(stderr,"subpixelize: width/height error\n");
    return 0; //failure
  }

  // skip white space and comments
  // if it starts with a digit, its the
  // line we're looking for
  if( fgets(b, 1023, stdin) == NULL ) {
    return 0; // failure
  }
  while( !isdigit(b[0]) ) {
    if( fgets(b, 1023, stdin) == NULL ) {
      return 0; //failure
    }
  }

  // get number of colors
  sscanf(b, "%u", &colors);
  if( colors > 255) {
    fprintf(stderr,"subpixelize: color error\n");
    return 0; //failure
  }

  // final image will be 3 times narrower
  sw = width / 3.0;
  sw3 = 3 * sw;
  
  // we write a slightly modified header to stdout
  sprintf(b, "P6\n%u %u\n%u\n", sw, height, colors);
  if( fwrite(b, strlen(b), 1, COMMAND) <= 0) {
    fprintf(stderr,"subpixelize: header error\n");
    return 0; //failure, probably broken pipe
  }

  // now we get a raster of width*height pixels with 3 bytes per pixel.
  w3 = 3 * width;  
  irasterbuf = (char *)malloc(w3);
  orasterbuf = (char *)malloc(sw3);
  for(i = 0; i < height; i++) {
    if( fread(irasterbuf, w3, 1, stdin) <= 0) {
      fprintf(stderr, "subpixelize: scanline read error\n");
      return 0; // failure, probably eof
    }
    (*render_subpixels)(irasterbuf,orasterbuf,sw);
    if( fwrite(orasterbuf, sw3, 1, COMMAND) <= 0) {
      fprintf(stderr, "subpixelize: scanline write error\n");
      return 0; // failure, probably broken pipe
    }
  }
  fflush(COMMAND); // not really needed, since we pclose immediately

  free(irasterbuf);
  free(orasterbuf);
  return 1; // success
}

// periodically check if window exists and stdin is not eof
// in case of failure, we die
void catch_alarm(int sig_num) {

  XGetWindowAttributes(dpy,win,&attributes_return);
  if( feof(stdin) ) {
    exit(1);
  } else {
    alarm(TIMEOUT);
  }
}

// reads a sequence of ppm images on stdin
// and pipes them to the command on the command line
// if the GHOSTVIEW variable is set, sends
// an X11 event after each ppm image is processed
int main(int argc, char **argv) {

  // input/output variables
  char buf[1024], pcommand[1024];
  int n, i, j, k;
  FILE *COMMAND;
  fd_set iset;
  int retval;
  struct timeval tv;
  FILE *ghost;
  void (*renderfn)();
  int success;

  /********************************
   * select rendering function    *
   ********************************/

  if( argc < 2) {
    usage(argv[0]);
  }

  i = 2; // index of command line
  if( strcmp(argv[1],"aa") == 0 ) {
    renderfn = render_subpixels_aa;
  } else if( strcmp(argv[1],"spr2") == 0 ) {
    renderfn = render_subpixels_spr2;
  } else if( strcmp(argv[1],"mat") == 0 ) {
    renderfn = render_subpixels_matrix;
    if( argc < 30 ) {
      usage(argv[0]);
    } 
    // read in the parameters
    for(; i < 29; i++) {
      *((double *)a + i - 2) = atof(argv[i]);
    }
    // recompute the sums
    for(j = 0; j < 3; j++) {
      asum[j] = 0;
      for(k = 0; k < 9; k++) {
	asum[j] += a[j][k];
      }
    }
  } else if( strcmp(argv[1],"spr") == 0 ) {
    renderfn = render_subpixels_spr;
  } else { // default
    renderfn = render_subpixels_spr;
  }

  /********************************
   * create command line and check*
   * for GHOSTVIEW environment    *
   ********************************/

  pcommand[0] = 0;
  for(; i < argc; i++) {
    strcat(pcommand, argv[i]);
    strcat(pcommand, " ");
  }
  if( !strlen(pcommand) ) {
    usage(argv[0]);
  }

  // read window id
  window_id = getenv("GHOSTVIEW");
  if( window_id != 0) {
    /********************************
     * set up X11 interface         *
     ********************************/

    // open the X11 connection
    dpy = XOpenDisplay((char *)NULL);
    assert( dpy != 0);

    scr = DefaultScreenOfDisplay(dpy);

    dest = 0;
    sscanf(window_id, "%ld %ld", &win, &dest);
    assert( dest == 0 );

    attributes.override_redirect = True;
    mwin = XCreateWindow(dpy, RootWindowOfScreen(scr),
			 0, 0, 1, 1, 0, CopyFromParent,
			 CopyFromParent, CopyFromParent,
			 CWOverrideRedirect, &attributes);
    NEXT = XInternAtom(dpy, "NEXT", False);
    PAGE = XInternAtom(dpy, "PAGE", False);
    DONE = XInternAtom(dpy, "DONE", False);

    signal(SIGALRM, catch_alarm); // for XNextEvent() below
  }

  /***********************************
   * process stdin and pass to stdout*
   ***********************************/

  success = 1;
  // each call to pipe_ppm() must properly match
  // a full PPM image. If something goes wrong,
  // we'll exit(1).
  // after each ppm image, if gv is running, we 
  // wait for an X11 event using XNextEvent().
  while( !feof(stdin) && success) {

    if( feof(stdin) ) { break; }
    // open a pipe to the command
    COMMAND = popen(pcommand, "w");
    success = pipe_ppm(COMMAND, renderfn);
    pclose(COMMAND);

    // tell gv that we've finished rendering
    // this gives mouse control back to the user
    // then we block until gv sends us an event
    // in case we wait for ever, we arrange for 
    // an alarm to verify that gv is still around.
    if( window_id != 0) {
      event.xclient.type = ClientMessage;
      event.xclient.display = dpy;
      event.xclient.window = win;
      event.xclient.message_type = PAGE;
      event.xclient.format = 32;
      event.xclient.data.l[0] = mwin;
      event.xclient.data.l[1] = dest;
      XSendEvent(dpy,win,False,0,&event);

      alarm(TIMEOUT);
      XNextEvent(dpy,&event);
      alarm(0);

    }
  } // end while


  if( window_id != 0 ) {
    /***********************************
     * finish and close X11 connection *
     ***********************************/

    event.xclient.type = ClientMessage;
    event.xclient.display = dpy;
    event.xclient.window = win;
    event.xclient.message_type = DONE;
    event.xclient.format = 32;
    event.xclient.data.l[0] = mwin;
    event.xclient.data.l[1] = dest;
    XSendEvent(dpy,win,False,0,&event);

    // close X11 connection
    XCloseDisplay(dpy);
  }

}

